import numpy as np

from .. import Operation as opmod 
from ..Operation import Operation

class WindowZip(Operation):
    """
    From input sequences for x and y, 
    produce an n-by-2 array 
    where x is bounded by the specified limits 
    """

    def __init__(self):
        input_names = ['x','y','x_min','x_max']
        output_names = ['x_window','y_window','x_y_window']
        super(WindowZip,self).__init__(input_names,output_names)        
        self.input_type['x'] = opmod.workflow_item
        self.input_type['y'] = opmod.workflow_item
        self.input_doc['x'] = 'array of x values'
        self.input_doc['y'] = 'array of y values'
        self.input_doc['x_min'] = 'inclusive minimum x value of output'
        self.input_doc['x_max'] = 'inclusive maximum x value of output'
        self.output_doc['x_window'] = 'array of x for x_min <= x <= x_max'
        self.output_doc['y_window'] = 'array of y for x_min <= x <= x_max'
        self.output_doc['x_y_window'] = 'n-by-2 array with x, y pairs for x_min <= x <= x_max'

    def run(self):
        xvals = self.inputs['x']
        yvals = self.inputs['y']
        x_min = self.inputs['x_min']
        x_max = self.inputs['x_max']
        idx_keep = ((xvals >= x_min) & (xvals <= x_max))
        x_y_window = np.array(zip(xvals[idx_keep],yvals[idx_keep]))
        self.outputs['x_window'] = x_y_window[:,0]
        self.outputs['y_window'] = x_y_window[:,1]
        self.outputs['x_y_window'] = x_y_window




import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation
from ... import optools

class BuildListFromBatch(Operation):
    """
    Given a batch output (list of dicts)
    and an output name (dict key), 
    harvest a list of outputs from the batch.
    """

    def __init__(self):
        input_names = ['batch_output','output_name']
        output_names = ['data_list']
        super(BuildListFromBatch,self).__init__(input_names,output_names)        
        self.input_doc['batch_output'] = 'list of dicts produced by a batch execution'
        self.input_doc['output_name'] = 'name of workflow output to be harvested'
        self.input_type['batch_output'] = opmod.workflow_item
        self.output_doc['data_list'] = 'list of the data fetched from batch_output'

    def run(self):
        b_out = self.inputs['batch_output']
        k = self.inputs['output_name']
        self.outputs['data_list'] = [d[k] for d in b_out]




from collections import OrderedDict
import copy

from ...Operation import Operation
from ... import Operation as opmod 
from ... import optools

class BatchFromFiles(Operation):
    """
    Take a list of file paths and use them as inputs
    for the repeated execution of a specified Workflow.
    Specify, by workflow uri, where this file path will be fed to the workflow.
    Collect outputs from the Workflow for each of the input files.
    """

    def __init__(self):
        input_names = ['file_list','workflow','input_name','extra_input_names','extra_inputs']
        output_names = ['batch_inputs','batch_outputs']
        super(BatchFromFiles,self).__init__(input_names,output_names)
        self.input_doc['file_list'] = 'list of file paths'
        self.input_doc['workflow'] = 'the Workflow to be executed'
        self.input_doc['input_name'] = 'name of the workflow input '\
        'where the file paths will be used'
        self.input_doc['extra_input_names'] = 'list of names '\
        'of batch workflow inputs to be set to extra_inputs '\
        'before batch-execution'
        self.input_doc['extra_inputs'] = 'data items '\
        'to be set to batch workflow inputs indicated by extra_input_names'
        self.output_doc['batch_inputs'] = 'list of dicts of [input_name:input_value]'
        self.output_doc['batch_outputs'] = 'list of dicts of [output_name:output_value] for all Workflow outputs'
        self.input_type['workflow'] = opmod.entire_workflow
        self.inputs['extra_input_names'] = []
        self.inputs['extra_inputs'] = []
        
    def run(self):
        batch_list = self.inputs['file_list'] 
        inpname = self.inputs['input_name'] 
        wf = self.inputs['workflow'] 
        n_batch = len(batch_list)
        self.outputs['batch_inputs'] = [None for ib in range(n_batch)] 
        self.outputs['batch_outputs'] = [None for ib in range(n_batch)] 
        if self.data_callback: 
            self.data_callback('outputs.batch_inputs',[None for ib in range(n_batch)])
            self.data_callback('outputs.batch_outputs',[None for ib in range(n_batch)])
        inps = self.inputs['extra_input_names']
        vals = self.inputs['extra_inputs']
        # Load any additional inputs...
        if any(inps): 
            for inpnm,inpval in zip(inps,vals):
                wf.set_wf_input(inpnm,inpval)
        self.message_callback('STARTING BATCH')
        for i,filename in zip(range(n_batch),batch_list):
            inp_dict = OrderedDict() 
            inp_dict[inpname] = filename
            wf.set_wf_input(inpname,filename)
            self.message_callback('BATCH RUN {} / {}'.format(i,n_batch-1))
            wf.execute()
            out_dict = wf.wf_outputs_dict()
            self.outputs['batch_inputs'][i] = inp_dict
            self.outputs['batch_outputs'][i] = out_dict
            if self.data_callback: 
                self.data_callback('outputs.batch_inputs.'+str(i),inp_dict)
                self.data_callback('outputs.batch_outputs.'+str(i),copy.deepcopy(out_dict))
        self.message_callback('BATCH FINISHED')




"""
Integrate an image, given calibration parameters.

This module builds a PyFAI.AzimuthalIntegrator 
to integrate an input image to I(q,chi).
"""

import numpy as np
import pyFAI

from ... import Operation as opmod 
from ...Operation import Operation

class Integrate2d(Operation):
    """
    Input image data (ndarray) and a dict of calibration parameters 
    Return q, chi, I(q,chi) 
    """
    def __init__(self):
        input_names = ['image_data','poni_dict']
        output_names = ['q','chi','I_at_q_chi','I_at_q','q_I']
        super(Integrate2d,self).__init__(input_names,output_names)
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['poni_dict'] = str( 'dict of calibration parameters; '
        + 'minimally including keys dist, poni1, poni2, rot1, rot2, rot3, pixel1, pixel2, wavelength;'
        + 'optionally including keys fpolz, detector, splineFile; '
        + 'same specifications as pyFAI .poni format calibration parameters')
        self.input_type['image_data'] = opmod.workflow_item
        self.input_type['poni_dict'] = opmod.workflow_item
        self.output_doc['q'] = 'scattering vector magnitude in 1/Angstrom'
        self.output_doc['chi'] = 'azimuthal angle in degrees'     
        self.output_doc['I_at_q_chi'] = 'integrated intensity at q, chi'
        self.output_doc['I_at_q'] = 'chi-integrated intensity at q'
        self.output_doc['q_I'] = 'q and I_at_q zipped together into an n-by-2 numpy array'

    def run(self):
        img = self.inputs['image_data']
        pd = self.inputs['poni_dict']
        if img is None or pd is None:
            return
        p = pyFAI.AzimuthalIntegrator()
        p.setPyFAI(**pd)
        fpolz = pd['fpolz']
        # use a mask to screen negative pixels
        # mask should be 1 for masked pixels, 0 for unmasked pixels
        s = int(img.shape[0])
        msk = np.ones((s,s))*(img <= 0)
        I_q_chi, q, chi = p.integrate2d(img, 1000, mask=msk, polarization_factor=fpolz, unit='q_A^-1')
        I_at_q = np.sum(I_q_chi,axis=0)
        #q = q * 1E9
        # save results to self.outputs
        self.outputs['q'] = q
        self.outputs['chi'] = chi 
        self.outputs['I_at_q_chi'] = I_q_chi 
        self.outputs['I_at_q'] = I_at_q 
        self.outputs['q_I'] = np.array([q,I_at_q]).T




from collections import OrderedDict

from PySide import QtCore

from ..qt.QTreeSelectionModel import QTreeSelectionModel
from ..core.operations.OpManager import OpManager

class QOpManager(OpManager,QTreeSelectionModel):
    """
    A QTreeSelectionModel for interacting with TreeModel OpManager.
    """

    def __init__(self):
        super(QOpManager,self).__init__()

    def headerData(self,section,orientation,data_role):
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} operations available".format(self.n_ops())
        else:
            return super(QOpManager,self).headerData(section,orientation,data_role) 

    def flags(self,idx):
        d = self.get_data_from_index(idx)
        if isinstance(d,dict) and not idx.column()==0:
            # Don't allow user to control enable flag on categories, at least for now
            # TODO: Consider allowing user to enable/disable entire categories
            #return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsTristate
            return QtCore.Qt.NoItemFlags
        else:
            return super(QOpManager,self).flags(idx)

    # Override setData to handle operation enable/disable via check boxes. 
    def setData(self,idx,val,data_role):
        if idx.column() == 0:
            return super(QOpManager,self).setData(index,val,data_role)
        elif data_role == QtCore.Qt.CheckStateRole:
            itm = self.get_from_index(idx)
            op_uri = self.get_uri_of_index(idx)
            try:
                self.set_op_enabled(op_uri,bool(val))
            except Exception as ex:
                msg = str('Failed to enable Operation {}. '.format(op_uri)
                + 'Error message: {}'.format(ex.message))
                self.write_log(msg)
                return False
            self.set_flagged(itm,self.default_flags.keys()[idx.column()-1],val)
            self.dataChanged.emit(idx,idx)
            return True
        else:
            return super(QTreeSelectionModel,self).setData(index,val,data_role)





import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class SubtractMaximumBackground(Operation):
    """
    Subtract a background from a foreground, with scaling to prevent over-subtraction.
    Optionally, input an intensity error array, 
    and get an error estimate for the background-subtracted intensity. 

    Operation originally contributed by Amanda Fournier.
    """

    def __init__(self):
        input_names = ['q_I', 'q_I_bg', 'dI', 'dI_bg']
        output_names = ['q_I', 'dI', 'bg_factor']
        super(SubtractMaximumBackground, self).__init__(input_names, output_names)
        self.input_doc['q_I'] = 'n-by-2 array of q values and corresponding intensity values'
        self.input_doc['q_I_bg'] = 'n-by-2 array, background corresponding to q_I'
        self.input_doc['dI'] = '1d array, error estimate of I (optional, default None)' 
        self.input_doc['dI_bg'] = '1d array, error estimate of I_bg (optional, default None)'
        self.output_doc['q_I'] = 'n-by-2 array of q values and background-subtracted intensity: I-(bg_factor*I_bg)'
        self.output_doc['dI'] = 'error estimate of background-subtracted intensity'
        self.output_doc['bg_factor'] = 'the factor the background was multiplied by '\
        ' before subraction to ensure positive values for output intensity'
        self.input_type['q_I'] = opmod.workflow_item
        self.input_type['q_I_bg'] = opmod.workflow_item

    def run(self):
        q_I = self.inputs['q_I']
        q_I_bg = self.inputs['q_I_bg']
        if q_I is None or q_I_bg is None:
            return 
        if not all(q_I[:,0] == q_I_bg[:,0]):
            msg = 'SPECTRUM AND BACKGROUND ON DIFFERENT q DOMAINS'
            raise ValueError(msg)
        I = q_I[:,1]
        I_bg = q_I[:,1]
        bad_data = (I < 0) | (I_bg <= 0) | np.isnan(I) | np.isnan(I_bg)
        bg_factor = np.min(I[~bad_data] / I_bg[~bad_data])
        I_out = I-(bg_factor*I_bg)
        dI = self.inputs['dI']
        dI_bg = self.inputs['dI_bg']
        dI_out = None
        if dI_bg is not None and dI is not None:
            dI_out = (dI**2+(bg_factor*dI_bg)**2)**0.5
        self.outputs['q_I'] = np.array(zip(q_I[:,0],I_out))
        self.outputs['dI'] = dI_out
        self.outputs['bg_factor'] = bg_factor
 



import numpy as np

from .. import Operation as opmod 
from ..Operation import Operation

class NoiseArray(Operation):
    """Creates and outputs a square array of noise"""

    def __init__(self):
        input_names = ['size']
        output_names = ['array']
        super(NoiseArray,self).__init__(input_names,output_names) 
        self.input_doc['size'] = 'dimension of output array'
        self.output_doc['array'] = 'an array of noise'
        self.inputs['size'] = 100 
    
    def run(self):
        s = self.inputs['size']
        self.outputs['array'] = np.random.rand(s,s) 




import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class TextureFeatures(Operation):
    """
    Analyzes the texture of an integrated diffractogram
    (q, chi, and I(q,chi)).

    Created on Mon Jun 06 2016.
    
    Originally contributed by Fang Ren.
    Citation:
    Fang Ren, et al.
    ACS Comb. Sci., 2017, 19(6), pp 377-385.
    """

    def __init__(self):
        input_names = ['q','chi','I']
        output_names = ['q_texture','texture','int_sqr_texture']
        super(TextureFeatures,self).__init__(input_names,output_names)
        self.input_doc['q'] = '1d array of momentum transfer values'
        self.input_doc['chi'] = '1d array of out-of-plane diffraction angles'
        self.input_doc['I'] = '2d array representing intensities at q,chi points'
        self.input_type['q'] = opmod.workflow_item
        self.input_type['chi'] = opmod.workflow_item
        self.input_type['I'] = opmod.workflow_item 
        self.output_doc['q_texture'] = 'q values at which the texture is analyzed'
        self.output_doc['texture'] = 'quantification of texture for each q'
        self.output_doc['int_sqr_texture'] = 'integral over q of the texture squared'

    def run(self):
        q = self.inputs['q']
        chi = self.inputs['chi']
        I = self.inputs['I']
        if q is None or chi is None or I is None:
            return
        q, chi = np.meshgrid(q,chi*np.pi/float(180))
        keep = (self.inputs['I'] != 0)
        I = keep.astype(int) * self.inputs['I']
        # TODO: This appears to be a binning operation.
        # Maybe the bin size should not be hard-coded. 
        I_sum = np.bincount((q.ravel()*100).astype(int), I.ravel().astype(int))
        count = np.bincount((q.ravel()*100).astype(int), keep.ravel().astype(int))
        I_ave = np.array(I_sum)/np.array(count)
        texsum = np.bincount((q.ravel()*100).astype(int), (I*np.cos(chi)).ravel())
        chi_count = np.bincount((q.ravel()*100).astype(int), (keep*np.cos(chi)).ravel())
        texture = np.array(texsum) / np.array(I_ave) / np.array(chi_count) - 1
        # TODO: remove this hard code
        step = 0.01
        q_texture = np.arange(step,np.max(q)+step,step)
        tsqr_int = np.nansum(texture ** 2)/float(q_texture[-1]-q_texture[0])
        self.outputs['q_texture'] = q_texture
        self.outputs['texture'] = texture
        self.outputs['int_sqr_texture'] = tsqr_int 






import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class CSVToArray(Operation):
    """
    Read a csv-formatted file into a numpy array.
    """

    def __init__(self):
        input_names = ['path']
        output_names = ['array']
        super(CSVToArray, self).__init__(input_names, output_names)
        self.input_doc['path'] = "path to .csv file"
        self.output_doc['array'] = "numpy array built from csv file contents"

    def run(self):
        path = self.inputs['path']
        if path is None:
            return
        self.outputs['array'] = np.loadtxt(path, dtype=float, delimiter=',')





import numpy as np

import yaml
from ... import Operation as opmod 
from ...Operation import Operation

class LoadYAML(Operation):
    """
    Load a YAML file, save the output of yaml.load(open(file_path,'r')) 
    """

    def __init__(self):
        input_names = ['file_path']
        output_names = ['yaml_output']
        super(LoadYAML, self).__init__(input_names, output_names)
        self.input_doc['file_path'] = 'path to YAML-formatted data file'
        self.output_doc['yaml_output'] = 'the result of yaml.load(file_path)' 

    def run(self):
        p = self.inputs['file_path']
        if p is None:
            return
        f = open(p,'r')
        ds = yaml.load(f)
        f.close()
        self.outputs['yaml_output'] = ds 





"""This module defines a class that presents an API for paws."""
from __future__ import print_function
import os
from functools import partial
from collections import OrderedDict
import re
import importlib

import yaml

from ..core import pawstools
from ..core import operations as ops
from ..core import plugins 
from ..core.operations import Operation as opmod
from ..core.operations import optools
from ..core.operations.OpManager import OpManager 
from ..core.workflow.WfManager import WfManager 
from ..core.plugins.PluginManager import PluginManager 

def start():
    """
    Instantiate and return a PawsAPI object. 

    paws.api.start() calls the PawsAPI constructor.

    :returns: a PawsAPI object
    :return type: paws.api.PawsAPI 
    """
    return PawsAPI()

class PawsAPI(object):
    """
    A container to facilitate interaction with a set of paws objects:
    an Operations Manager, a Workflow Manager, and a Plugins Manager. 
    """

    def __init__(self):
        super(PawsAPI,self).__init__()
        # Assign a function(str) to PawsAPI.logmethod
        # to change where messages get printed
        self._op_manager = OpManager()
        self._plugin_manager = PluginManager()
        self._wf_manager = WfManager()
        self._wf_manager.plugin_manager = self._plugin_manager

        # TODO: load_cats and load_ops should happen outside the api.__init__
        # so that different api instances can have different operations loaded
        self._op_manager.load_cats(ops.cat_list) 
        self._op_manager.load_ops(ops.cat_op_list)
        self._current_wf_name = None 
        self.logmethod = print

    def set_logmethod(self,lm):
        """
        Sets the logmethod, which is the function
        that is called to handle messages.

        Parameters
        ----------
        lm : function
            function to be called for logging messages
        """
        self.logmethod = lm
        self._op_manager.logmethod = lm
        self._plugin_manager.logmethod = lm
        self._wf_manager.logmethod = lm

    def info(self):   
        info_msg = str('PAWS: the Platform for Automated Workflows by SSRL. '
        + 'Version: {}'.format(pawstools.version))
        print(info_msg)
        return info_msg
 
    def activate_op(self,op_uri):
        """
        Import the Operation indicated by op_uri, and tag it as active.
        The Operation becomes available to add to workflows via paws.api.add_op()
        """
        self._op_manager.set_op_enabled(op_uri)

    def deactivate_op(self,op_uri):
        """
        Disable the Operation indicated by op_uri.
        The Operation cannot be added to Workflows until it is enabled again. 
        """
        self._op_manager.set_op_enabled(op_uri,False)

    def enable_op(self,op_tag,wfname=None):
        self.get_wf(wfname).set_op_enabled(op_tag,True)

    def disable_op(self,op_tag,wfname=None):
        self.get_wf(wfname).set_op_enabled(op_tag,False)

    def enable_plugin(self,pgin_name=''):
        """
        This tests the compatibility between the environment and the named plugin
        by attempting to import the plugin.
        If this does not throw an ImportError, 
        then the environment satisfies the plugin dependencies.
        """
        pkg = plugins.__name__
        mod = importlib.import_module('.'+pgin_name,pkg)

    def select_wf(self,wfname):
        """
        Sets the current workflow for the API instance.
        This is only to simplify subsequent api calls:
        anywhere there is an optional workflow name input,
        the default behavior is to apply the call to the current workflow.
        """
        if wfname in self._wf_manager.workflows.keys():
            self._current_wf_name = wfname
        else:
            msg = str('requested workflow {} not found in {}'
            .format(wfname,self._wf_manager.workflows.keys()))
            raise ValueError(msg)

    def n_wf(self):
        return self._wf_manager.n_wf()

    def current_wf_name(self):
        return self._current_wf_name

    def current_wf(self):
        if self._current_wf_name:
            return self._wf_manager.workflows[self._current_wf_name]    
        else:
            return None

    def get_wf(self,wfname=None):
        if wfname is None:
            return self.current_wf()
        else:
            return self._wf_manager.workflows[wfname]

    def get_op(self,opname,wfname=None):
        return self.get_wf(wfname).get_data_from_uri(opname) 

    def add_op(self,op_tag,op_spec,wfname=None):
        if ops.load_flags[op_spec]:
            wf = self.get_wf(wfname)
            # get the op referred to by op_spec
            op = self._op_manager.get_data_from_uri(op_spec)
            # instantiate with default inputs
            op = op()
            op.load_defaults()
            wf.add_op(op_tag,op)
        else:
            msg = str('Attempted to add Operation {}, '.format(op_spec)
            + 'but this Operation has not been activated. '
            + 'Activate it with paws.api.activate_op() '
            + 'before adding it to a workflow.')
            self.logmethod(msg) 
            raise pawstools.OperationDisabledError(msg)

    def add_wf(self,wfname):
        """
        Adds a workflow to the workflow manager.
        Input the workflow name.
        If no current workflow is selected,
        calls self.select_wf(wfname) at the end,
        selecting the new workflow for subsequent api calls.
        """
        self._wf_manager.add_wf(wfname)
        if not self._current_wf_name:
            self.select_wf(wfname)

    def add_plugin(self,pgin_tag,pgin_name):
        pgin = self._plugin_manager.load_plugin(pgin_name)
        self._plugin_manager.add_plugin(pgin_tag,pgin())

    def load_plugin(self,pgin_module):
        return self._plugin_manager.load_plugin(pgin_module)

    def set_plugin_input(self,pgin_tag,input_name,val=None,tp=None):
        pgin = self._plugin_manager.get_data_from_uri(pgin_tag)
        if tp is None:
            # if type is not specified, take the plugin's default 
            tp = pgin.input_type[input_name]
        else:
            if tp in opmod.input_types:
                tp = opmod.valid_types[ opmod.input_types.index(tp) ]
        # TODO: separate opmod from plugin stuff
        if tp == opmod.no_input or val is None: 
            pgin.inputs[input_name] = None
        elif tp == opmod.basic_type:
            pgin.inputs[input_name] = val
        else:
            msg = '[{}] failed to parse plugin input {}, tp: {}, val: {}'.format(
            __name__,input_name,tp,val)
            raise ValueError(msg)

    def start_plugin(self,pgin_name):
        pgin = self.get_plugin(pgin_name)
        pgin.start()

    def get_plugin(self,pgin_name):
        return self._plugin_manager.get_plugin(pgin_name)

    def remove_op(self,op_tag,wfname=None):
        wf = self.get_wf(wfname)
        wf.remove_item(op_tag)

    def add_wf_input(self,wf_input_name,input_uris,wfname=None):
        """
        Add an input to the workflow specified by wfname,
        and specify its workflow routing by any number of input_uris,
        which should refer to the inputs of operations in the workflow.
        When the workflow is asked to set this input to some value x,
        it will set all of the provided input_uris to x.
        """
        self.get_wf(wfname).connect_wf_input(wf_input_name,input_uris) 

    def add_wf_output(self,wf_output_name,output_uris,wfname=None):
        """
        Add an output to the workflow specified by wfname,
        and specify one or more output_uris for pieces of workflow data
        that will be referenced to this workflow output.
        If multiple output_uris are specified, they will be packed as a list.
        """
        self.get_wf(wfname).connect_wf_output(wf_output_name,output_uris) 

    def remove_wf_input(self,wf_input_name,wfname=None):
        self.get_wf(wfname).break_wf_input(wf_input_name) 

    def remove_wf_output(self,wf_output_name,wfname=None):
        self.get_wf(wfname).break_wf_output(wf_output_name) 

    def set_wf_input(self,wf_input_name,val=None,wfname=None):
        if wfname is None:
            wfname = self._current_wf_name
        self.get_wf(wfname).set_wf_input(wf_input_name,val) 

    def set_input(self,opname,input_name,val=None,tp=None,wfname=None):
        if wfname is None:
            wfname = self._current_wf_name
        op = self.get_op(opname,wfname)
        if not input_name in op.inputs.keys():
            msg = str('Input name {} not valid for Operation {} ({}).'
            .format(input_name,opname,type(op).__name__))
            raise KeyError(msg)
        if tp is None:
            # if type is not specified, take the operation's default 
            tp = op.input_type[input_name]
        elif tp in opmod.input_types:
            # type specified by string: convert to enum
            tp = opmod.valid_types[ opmod.input_types.index(tp) ]
        if not tp in opmod.valid_types:
            # tp is neither a string or an enum
            msg = '[{}] failed to parse input type: {}'.format(
            __name__,tp)
            raise ValueError(msg)
        elif tp == opmod.runtime_type:
            # set inputlocator.val to None, 
            # so that this object will NOT attempt
            # to be serialized when save_to_wfl() is called.
            il = opmod.InputLocator(tp,None)
        else:
            # all other types, val can be serialized,
            # so it is referenced in the InputLocator. 
            il = opmod.InputLocator(tp,val)
        op.input_locator[input_name] = il
        if tp == opmod.basic_type:
            self.get_wf(wfname).set_op_item(opname,'inputs.'+input_name,val)

    def get_input_data(self,opname,input_name,wfname=None):
        if wfname is None:
            wfname = self._current_wf_name
        op = self.get_op(opname,wfname) 
        if not input_name in op.inputs.keys():
            msg = str('Input name {} not valid for Operation {} ({}).'
            .format(input_name,opname,type(op).__name__))
            raise KeyError(msg)
        return op.inputs[input_name]

    def get_input_setting(self,opname,input_name,wfname=None):
        if wfname is None:
            wfname = self._current_wf_name
        op = self.get_op(opname,wfname) 
        if not input_name in op.inputs.keys():
            msg = str('Input name {} not valid for Operation {} ({}).'
            .format(input_name,opname,type(op).__name__))
            raise KeyError(msg)
        return op.input_locator[input_name].val

    def get_output(self,opname,output_name=None,wfname=None):
        op = self.get_op(opname,wfname)
        if output_name is not None:
            return op.outputs[output_name]
        else:
            return op.outputs

    def execute(self,wfname=None):
        if wfname is None:
            wfname = self._current_wf_name
        self._wf_manager.run_wf(wfname)
        
    def save_config(self):
        ops.save_config()

    def wfl_dict(self):
        d = {} 
        d['OP_ACTIVATION_FLAGS'] = ops.load_flags
        d['PAWS_VERSION'] = pawstools.version 
        wfman_dict = OrderedDict()
        for wfname,wf in self._wf_manager.workflows.items():
            wfman_dict[wfname] = wf.wf_setup_dict() 
        d['WORKFLOWS'] = wfman_dict
        pgin_dict = OrderedDict() 
        for pgin_name in self._plugin_manager.list_plugin_tags():
            pgin = self._plugin_manager.get_data_from_uri(pgin_name)
            pgin_dict[pgin_name] = self._plugin_manager.plugin_setup_dict(pgin)
        d['PLUGINS'] = pgin_dict
        return d

    def save_to_wfl(self,wfl_filename):
        """
        Save the current workflows and plugins
        to a .wfl (YAML) file,
        specified by wfl_filename.
        If the given filename does not have the .wfl extension,
        it will be appended.
        """
        self.logmethod( 'saving current state to {}'.format(wfl_filename) )
        if not os.path.splitext(wfl_filename)[1] == '.wfl':
            wfl_filename = wfl_filename + '.wfl'
        d = self.wfl_dict()
        #pawstools.update_file(wfl_filename,d)
        pawstools.save_file(wfl_filename,d)

    def load_from_wfl(self,wfl_filename):
        f = open(wfl_filename,'r')
        d = yaml.load(f)
        f.close()
        if 'PAWS_VERSION' in d.keys():
            wfl_version = d['PAWS_VERSION']
        else:
            wfl_version = '0.0.0'
        wfl_vparts = re.match(r'(\d+)\.(\d+)\.(\d+)',wfl_version)
        wfl_vparts = list(map(int,wfl_vparts.groups()))
        current_vparts = re.match(r'(\d+)\.(\d+)\.(\d+)',pawstools.version)  
        current_vparts = list(map(int,current_vparts.groups()))
        if wfl_vparts[0] < current_vparts[0] or wfl_vparts[1] < current_vparts[1]:
            # WARNING
            self.logmethod('WARNING: paws (version {}) '\
            'is trying to load a state built in version {} - '\
            'this is likely to cause things to crash, '\
            'until the workflows and plugins are reviewed/refactored '\
            'under the current version.'.format(pawstools.version,wfl_version))  
        if 'OP_ACTIVATION_FLAGS' in d.keys():
            for opname,flag in d['OP_ACTIVATION_FLAGS'].items():
                if opname in ops.load_flags.keys():
                    if ops.load_flags[opname]:
                        self.activate_op(opname)
        if 'WORKFLOWS' in d.keys():
            wf_dict = d['WORKFLOWS']
            for wfname,wfspec in wf_dict.items():
                self._wf_manager.load_from_dict(wfname,wfspec,self._op_manager)
                self.select_wf(wfname)
        if 'PLUGINS' in d.keys():
            pgin_dict = d['PLUGINS']
            for pgin_name,pgin_spec in pgin_dict.items():
                self._plugin_manager.load_from_dict(pgin_name,pgin_spec)
    
    def op_count(self,wfname=None):
        return self.get_wf(wfname).n_children()

    def list_wf_tags(self):
        return self._wf_manager.workflows.keys()
    
    def list_op_tags(self,wfname=None):
        return self.get_wf(wfname).list_op_tags()

    def list_plugin_tags(self):
        return self._plugin_manager.list_plugin_tags()




import numpy as np
import copy

from ... import Operation as opmod 
from ...Operation import Operation
from ... import optools
from ....tools.saxs import saxs_fit

class PostProcessTimeSeries(Operation):
    """Re-fit a set of SAXS spectra chronologically, to refine results."""

    def __init__(self):
        input_names = ['batch_outputs','time_key','qI_key','flags_key','params_key']
        output_names = ['new_outputs']
        super(PostProcessTimeSeries,self).__init__(input_names,output_names)        
        self.input_doc['batch_outputs'] = 'list of dicts '\
            'of workflow outputs, produced by a batch execution'
        self.input_doc['time_key'] = 'dict key for time stamp in batch_outputs'
        self.input_doc['flags_key'] = 'dict key for getting spectrum '\
            '(n-by-2 array of q and I) from batch_outputs'
        self.input_doc['flags_key'] = 'dict key for '\
            'population flags in batch_outputs'
        self.input_doc['params_key'] = 'dict key for '\
            'SAXS equation parameters in batch_outputs' 
        self.input_type['batch_outputs'] = opmod.workflow_item
        self.output_doc['new_outputs'] = 'list of dicts '\
            'containing time stamps and newly refined parameters'
        self.inputs['time_key'] = 'time'
        self.inputs['flags_key'] = 'flags'
        self.inputs['params_key'] = 'params'

    def run(self):
        b_out = self.inputs['batch_outputs']
        kq_I = self.inputs['qI_key']
        ktime = self.inputs['time_key']
        kflags = self.inputs['flags_key']
        kparams = self.inputs['params_key']
    
        t = [b[ktime] for b in b_out]
        f = [b[kflags] for b in b_out]
        p = [b[kparams] for b in b_out]
        q_I = [b[kq_I] for b in b_out]
        new_outputs = []

        n_batch = len(b_out)
        idx_batch = np.arange(0,n_batch)

        for idx in idx_batch:
            p_i = copy.deepcopy(p[idx])
            if not idx==0:
                p_test = copy.deepcopy(p[idx-1])
                p_opt_test = saxs_fit.fit_spectrum(q_I[idx],f[idx],p_test,p_i['fixed_params'],p_i['objfun']) 
                if p_opt_test['objective_after'] < p_i['objective_after']:
                    p_i = p_opt_test
            if not idx == n_batch-1:
                p_test = copy.deepcopy(p[idx+1])
                p_opt_test = saxs_fit.fit_spectrum(q_I[idx],f[idx],p_test,p_i['fixed_params'],p_i['objfun']) 
                if p_opt_test['objective_after'] < p_i['objective_after']:
                    p_i = p_opt_test
            new_outputs.append({ktime:t[idx],kparams:p_i})
        self.outputs['new_outputs'] = new_outputs

            # an alternate approach using statistics of several nearby points...
            #p_i = copy.deepcopy(p[idx])
            ## Collect params for several of the nearest points
            #idx_use = (abs(idx_batch-idx) < 4) & idx_batch != idx
            #p_stats = dict.fromkeys(p_i.keys())
            #for k in p_i.keys():
            #    p_stats[k] = []
            ## Collect params of any nearby, relevant samples 
            #for p_j in p[idx_use]:
            #    if k in p_j.keys():
            #        p_stats[k].append(p_j[k])
            ## Are params of this idx outlier-like?
            #for k in p_i.keys():
            #    p_mean = np.mean(p_stats)
            #    p_std = np.std(p_stats)
            #    if p_i[k] > p_mean+p_std or p_i[k] < p_mean-p_std:
            #        # If so, re-set to the local mean
            #        p_i[k] = p_mean
            ## Re-fit, then save the result
            #p_reopt = saxs_fit.fit_spectrum(q_I[idx],f[idx],p_i,pfix[idx])
            #p_new.append(p_reopt)






from collections import OrderedDict
import copy

import numpy as np
from scipy.optimize import curve_fit

from ... import Operation as opmod 
from ...Operation import Operation
from ....tools.saxs import saxs_fit

class SpectrumFit(Operation):
    """
    Use a measured SAXS spectrum (I(q) vs. q),
    to optimize the parameters of a theoretical SAXS spectrum
    for one or several populations of scatterers. 
    Works by minimizing an objective function that compares
    the measured spectrum against the theoretical result.
    TODO: document the algorithm here.

    Input arrays of q and I(q), 
    a string indicating choice of objective function, 
    a dict of features describing the spectrum,
    and a list of strings indicating which keys in the dict 
    should be used as optimization parameters.
    The input features dict includes initial fit parameters
    as well as the flags indicating which populations to include.
    The features dict is of the same format as
    SpectrumProfiler and SpectrumParameterization outputs.

    Outputs a return code and the features dict,
    with entries updated for the optimized parameters.
    Also returns the theoretical result for I(q),
    and a renormalized measured spectrum for visual comparison.
    """

    def __init__(self):
        input_names = ['q_I','flags','params','fixed_params']#,'objfun','constraints']
        output_names = ['params','q_I_opt']
        super(SpectrumFit, self).__init__(input_names, output_names)
        self.input_doc['q_I'] = 'n-by-2 array of wave vectors (1/Angstrom) and intensities (counts)'
        self.input_doc['flags'] = 'dict of flags indicating what populations to fit'
        self.input_doc['params'] = 'dict of initial values for the scattering equation parameters '\
            'for each of the populations specified in the input flags'
        self.input_doc['fixed_params'] = 'list of strings (keys) indicating '\
            'which parameters to hold fixed during optimization'
        #self.input_doc['objfun'] = 'string indicating objective function for optimization- '\
        #    'see documentation of tools.saxs.saxs_fit.fit_spectrum() for supported objective functions'
        #self.input_doc['constraints'] = 'list of strings indicating constraints- '\
        #    'see documentation of tools.saxs.saxs_fit.fit_spectrum() for supported constraints'
        self.output_doc['params'] = 'dict of scattering equation parameters copied from inputs, '\
            'with values optimized for all keys not listed in fixed_params'
        self.output_doc['q_I_opt'] = 'n-by-2 array of q and the optimized computed intensity spectrum'
        self.input_type['q_I'] = opmod.workflow_item
        self.input_type['flags'] = opmod.workflow_item
        self.input_type['params'] = opmod.workflow_item
        self.inputs['fixed_params'] = [] 

    def run(self):
        q_I = self.inputs['q_I']
        f = self.inputs['flags']
        p = self.inputs['params']
        fixkeys = self.inputs['fixed_params']
        if not f['bad_data']:
            p_opt = saxs_fit.fit_spectrum(q_I,f,p,fixkeys,'chi2log')
            I_opt = saxs_fit.compute_saxs(q_I[:,0],f,p_opt)
            q_I_opt = np.array([q_I[:,0],I_opt]).T
            self.outputs['params'] = p_opt 
            self.outputs['q_I_opt'] = q_I_opt




import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation
from ... import optools

class XYDataFromBatch(Operation):
    """
    Harvest two arrays from a batch output (a list of dicts). 
    Takes a batch output, a key for x values, and a key for y values.
    """

    def __init__(self):
        input_names = ['batch_output','x_key','y_key','x_shift_flag']
        output_names = ['x','y','x_y','x_y_sorted']
        super(XYDataFromBatch,self).__init__(input_names,output_names)        
        self.input_doc['batch_output'] = 'list of dicts produced by a batch execution.'
        self.input_doc['x_uri'] = 'uri of data for x. Must be in batch.saved_items().'
        self.input_doc['y_uri'] = 'uri of data for y. Must be in batch.saved_items().'
        self.input_doc['x_shift_flag'] = 'if True, shift x data so that its minimum value is zero.' 
        self.input_type['batch_output'] = opmod.workflow_item
        self.inputs['x_shift_flag'] = False
        self.output_doc['x'] = 'array of the x values in batch output order.'
        self.output_doc['y'] = 'array of the y values in batch output order.'
        self.output_doc['x_y'] = 'n-by-2 array of x and y values in batch output order.'
        self.output_doc['x_y_sorted'] = 'n-by-2 array of x and y values, sorted for increasing x.'

    def run(self):
        b_out = self.inputs['batch_output']
        kx = self.inputs['x_key']
        ky = self.inputs['y_key']
        if b_out is None or kx is None or ky is None:
            return
        #x_all = np.array([optools.get_uri_from_dict(kx,d) for d in b_out],dtype=float)
        #y_all = np.array([optools.get_uri_from_dict(ky,d) for d in b_out],dtype=float)
        x_list = []
        y_list = []
        for d in b_out:
            if kx in d and ky in d:
                x_list.append(d[kx])
                y_list.append(d[ky])
        #x_list = [d[kx] if kx in d and ky in d for d in b_out]
        #y_list = [d[ky] if kx in d and ky in d for d in b_out]
        #y_0 = b_out[0][kx]
        #x_0 = b_out[0][ky]
        x_all = np.array(x_list)
        y_all = np.array(y_list)
        x_len = len(x_list)
        if x_len > 0:
            xmin = np.min(x_all)
        else:
            xmin = 0
        if self.inputs['x_shift_flag']:
            x_all = x_all - xmin
            #xmin = 0 
        self.outputs['x'] = x_all 
        self.outputs['y'] = y_all
        self.outputs['x_y'] = np.array(zip(x_all,y_all))
        #self.outputs['x_y_sorted'] = np.sort(np.array(zip(x_all,y_all)),0)
        i_xsort = np.argsort(x_all)
        y_xsort = y_all[i_xsort]
        x_sort = x_all[i_xsort] 
        self.outputs['x_y_sorted'] = np.array(zip(x_sort,y_xsort))





from __future__ import print_function
import unittest
import os

import paws.api
import test_api
import test_op

runner = unittest.TextTestRunner(verbosity=3)

print('======================================================================')
print('--- testing paws.api ---'+os.linesep)
api_tests = unittest.TestSuite()
api_tests.addTest(test_api.TestAPI('test_start_stop'))
runner.run(api_tests)
print(os.linesep+'--- done testing paws.api ---')
print('======================================================================')

# Start an API to use in subsequent tests
paw = paws.api.start()

print('======================================================================')
print('--- testing activation of operations ---'+os.linesep)
# Test paw.activate_op for all Operations
activate_op_tests = unittest.TestSuite()
for op_uri in paw._op_manager.list_ops():
    activate_op_tests.addTest(test_op.TestOp('test_activate_op',op_uri,paw))
activate_op_result = runner.run(activate_op_tests)
print(os.linesep+'--- done testing activation of operations ---')
print('======================================================================')

# Make a list of the Operations that could be activated
ops_skipped = [t[0].op_uri for t in activate_op_result.skipped]
ops_active = [opname for opname in paw._op_manager.list_ops() if opname not in ops_skipped]

print('======================================================================')
print('--- testing active Operations ---'+os.linesep)
# Test instantiation for all active Operations
op_tests = unittest.TestSuite()
for op_uri in ops_active:
    op_tests.addTest(test_op.TestOp('test_op',op_uri,paw))
op_result = runner.run(op_tests)
print('======================================================================')
op_run_tests = unittest.TestSuite()
for op_uri in ops_active:
    op_run_tests.addTest(test_op.TestOp('test_run',op_uri,paw))
op_run_result = runner.run(op_run_tests)
print(os.linesep+'--- done testing active Operations ---')
print('======================================================================')

# TODO: Test plugins

print('======================================================================')
print('--- testing api for workflows ---'+os.linesep)
api_tests = unittest.TestSuite()
api_tests.addTest(test_api.TestAPI('test_add_wf',paw))
api_tests.addTest(test_api.TestAPI('test_add_op',paw))
api_tests.addTest(test_api.TestAPI('test_execute',paw))
api_tests.addTest(test_api.TestAPI('test_save',paw))
api_tests.addTest(test_api.TestAPI('test_load',paw))
runner.run(api_tests)
print(os.linesep+'--- done testing api for workflows ---')
print('======================================================================')





import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class EasyZingers1d(Operation):
    """
    This Operation attempts to remove zingers 
    from 1d spectral data (I(q) versus q).
    Zingers are replaced with the average intensity
    in a window around where the zinger was found.
    """

    def __init__(self):
        input_names = ['q_I']
        output_names = ['q_I_dz','zmask']
        super(EasyZingers1d, self).__init__(input_names, output_names)
        self.input_doc['q_I'] = 'n-by-2 array of q values and corresponding intensities'
        self.output_doc['q_I_dz'] = 'same as input q_I but with zingers removed'
        self.output_doc['zmask'] = 'array of booleans, same shape as q, true if there is a zinger at q, else false'
        self.input_type['q_I'] = opmod.workflow_item

    def run(self):
        q_I = self.inputs['q_I']
        if q_I is None:
            return
        q = q_I[:,0]
        I = q_I[:,1]
        I_dz = np.array(I)
        zmask = np.array(np.zeros(len(q)),dtype=bool)
        w = 10
        I_ratio_limit = 10
        idx_z = []
        I_z = []
        stop_idx = len(q)-w-1
        test_range = iter(range(w,stop_idx))
        idx = test_range.next()
        while idx < stop_idx-1:
            Ii_l = np.array(I[idx-w:idx+1])
            Ii_r = np.array(I[idx:idx+w+1])
            Iimin_l = np.min(Ii_l)
            Iimin_r = np.min(Ii_r)
            Ii_l = Ii_l - Iimin_l
            Ii_r = Ii_r - Iimin_r
            Imean_l = np.mean(Ii_l[:-1]) 
            Imean_r = np.mean(Ii_r[1:]) 
            if not Imean_l == 0 and not Imean_r == 0:
                I_ratio_l = Ii_l[-1]/Imean_l
                I_ratio_r = Ii_r[0]/Imean_r
                if I_ratio_l > I_ratio_limit or I_ratio_r > I_ratio_limit:
                    idx_z.append(idx)
                    zmask[idx] = True
                    I_dz[idx] = np.nan
            idx = test_range.next() 
        q_z = [q[i] for i in idx_z]
        I_z = [I[i] for i in idx_z]
        newIvals = np.zeros(len(q_z)) 
        for qi,zi in zip(idx_z,range(len(idx_z))):
            Idzi = I_dz[qi-w:qi+w+1]
            Idzi = Idzi[~np.isnan(Idzi)]
            newIvals[zi] = np.mean(Idzi)
        for i,iq in zip(range(len(idx_z)),idx_z):
            I_dz[iq] = newIvals[i]
        self.outputs['q_I_dz'] = np.array(zip(q,I_dz))
        self.outputs['zmask'] = zmask

        #if any(idx_z):
        #if True:
        #    from matplotlib import pyplot as plt
        #    plt.figure(1)
        #    plt.semilogy(q,I)
        #    plt.semilogy(q,I_dz,'g')
        #    for i in idx_z:
        #        plt.semilogy(q[i],I[i],'ro')
        #    plt.show()
            




from __future__ import print_function
import socket 

from .PawsPlugin import PawsPlugin
from ..operations import Operation as opmod

class SpecClientPlugin(PawsPlugin):

    def __init__(self):
        input_names = ['host','port']
        super(SpecClientPlugin,self).__init__(input_names)
        self.input_doc['host'] = 'string representing host name or IP address'
        self.input_doc['port'] = 'integer port number at which SpecInfoServer should be listening on host' 
        self.history = []
        self.cmd_list = []
       
    def start(self):
        self.host = self.inputs['host'] 
        self.port = self.inputs['port'] 
        try:
            self.sock = socket.create_connection((self.host, self.port)) 
        except:
            self.sock = None
            print("Spec info server connection problem: {}".format(socket.error))
            pass

    def stop(self):
        print("PawsSpecClient stop")
        self.sock.close()
        # TODO: Anything else for cleaning up the connections.

    def description(self):
        desc = str('SpecInfoServer Client Plugin: '
            + 'This is a TCP Client used to communicate with SpecInfoServer. '
            + 'Startup requires a host name and a port number, '
            + 'where it is expected that SpecInfoServer will be listening.')
        return desc

    def content(self):
        return {'inputs':self.inputs, 'history':self.history, 'socket':self.sock}

    def receiveLine(self):
        # Handle lines received from Spec server
        buffer = bytearray(b' ' * 1024) 
        try:
            len = self.sock.recv_into(buffer) 
        except:
            print("Spec info server connection problem: {}".format(socket.error))
            return "ERRCOMM"

        buffer = buffer.strip(' ')
        return buffer
    
    def sendLine(self, line):
        try:
            self.sock.sendall(bytearray(line))
        except:
            print("Spec info server connection problem: {}".format(socket.error))
            self.history.append('NO CONNECTION TO SPEC INFO SERVER')

    def sendCmd(self, cmd):
        reply = ""
        # Will need timeout
        while reply == "" or reply.startswith("spec is busy!"):
            self.history.append('SEND: {}'.format(cmd))
            self.sendLine(cmd)
            reply = self.receiveLine()
        self.history.append('COMMAND: {} RESPONSE: {}'.format(cmd,reply))

    def send_commands(self, cmd_list):
        self.cmd_list = cmd_list
        if self.cmd_list == []:
            self.history.append('COMMAND LIST EMPTY')
        else:
            for s in self.cmd_list:
                self.sendCmd(s)

    def send_text(self, txt):
        self.cmd_list = [txt]
        send_commands(self.cmd_list)

#--------------------------------------------------------------------------------------------------------------------------#
#    p.addCommand("!cmd slacx_mar_data_path = 'my_mar_data'")
#    p.addCommand("!cmd slacx_pd_filename = 'my_pd_filename'")
#    p.addCommand("!cmd slacx_loopscan_npoints = 2")
#    p.addCommand("!cmd slacx_loopscan_counting_time = 2")
#    p.addCommand("!cmd runme")
#    p.addCommand("?sta")



import numpy as np
from PIL import Image

from ... import Operation as opmod 
from ...Operation import Operation

class LoadTif_PIL(Operation):
    """
    Takes a filesystem path that points to a .tif,
    outputs image data and metadata from the file. 
    """

    def __init__(self):
        input_names = ['file_path']
        output_names = ['image_data','metadata']
        super(LoadTif_PIL,self).__init__(input_names,output_names)
        self.input_doc['file_path'] = 'path to a .tif image'
        self.output_doc['image_data'] = '2D array representing pixel values'
        self.output_doc['metadata'] = 'Dictionary of image metadata'
        
    def run(self):
        p = self.inputs['file_path']
        if p is None:
            return
        pil_img = Image.open(p)
        self.outputs['image_data'] = np.array(pil_img)
        self.outputs['metadata'] = pil_img.info

